import xesam
import xesam.query
from xesam import *

import dbus
import dbus.glib
import dbus.service
import gobject

from Queue import Empty
from Queue import Queue

class UnstartedSearchException (Exception):
	"""
	Throw if you try and retrieve data from an unstarted search
	"""
	def __init__ (self, msg):
		Exception.__init__ (msg)

class GenericServer (dbus.service.Object):
	
	def __init__ (self, bus_name, object_path):
		_real_bus_name = dbus.service.BusName(bus_name, dbus.SessionBus())
		
		dbus.service.Object.__init__ (self, _real_bus_name, object_path)
		
		self.bus_name = _real_bus_name
		self.object_path = object_path
		
		_bus = dbus.SessionBus()
		_proxy_obj = _bus.get_object("org.freedesktop.DBus", "/org/freedesktop/DBus")
		self.dbus = dbus.Interface (_proxy_obj, "org.freedesktop.DBus")
		
		self.mainloop = None
	
	def start (self):
		self.mainloop = gobject.MainLoop()
		try:
			self.mainloop.run()
		except 	KeyboardInterrupt:
			xesam.debug ("Caught keyboard interrupt. Shutting down.")
			raise SystemExit
	
	def stop (self):
		self.mainloop.quit()

class Searchable (gobject.GObject):

	
	__gsignals__ = {
		"hits-added" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_STRING, gobject.TYPE_INT]),
		"hits-removed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_STRING, gobject.TYPE_PYOBJECT]),
		"hits-modified" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_STRING, gobject.TYPE_PYOBJECT]),
		"state-changed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_PYOBJECT]),
		"search-done" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_STRING]),
	}
	
	def __init__ (self):
		gobject.GObject.__init__ (self)
		
	def NewSession (self):
		"""
		Sender should be the name of the dbus object doing the request
		Fx ":1.24"
		"""
		raise NotImplementedError
	
	def SetProperty (self, session, name, value):
		raise NotImplementedError
	
	def GetProperty (self, session, name):
		raise NotImplementedError
	
	def CloseSession (self, session):
		raise NotImplementedError
	
	def NewSearch (self, session, query_xml):
		raise NotImplementedError
	
	def StartSearch (self, search):
		raise NotImplementedError	
	
	def GetHitCount (self, search):
		raise NotImplementedError
	
	def GetHits (self, search, count):
		raise NotImplementedError
	
	def GetHitData (self, search, hit_ids, properties):
		raise NotImplementedError
	
	def CloseSearch (self, search):
		raise NotImplementedError
	
	def GetState (self):
		raise NotImplementedError
	
	def HitsAdded (self, search, count):
		pass
		
	def HitsRemoved (self, search, hit_ids):
		pass
		
	def HitsModified (self, search, hit_ids):
		pass
	
	def StateChanged (self, state_info):
		pass
	
	def SearchDone (self):
		pass

class HandleFactory :
	"""
	Interface for creating and handling both session and search handles.
	In the default implementation session handles are just integers in
	string form, and search handles areon the form
		
		<session_num>:<search_num>
	
	The contract of this interface is that each handle, both session
	and search, must be globally unique.
	"""
	def __init__ (self):
		self._session_counter = "0"
		self._search_counter = "0"
	
	def new_session_handle (self):
		self._session_counter = str(int(self._session_counter) + 1)
		return self._session_counter
	
	def new_search_handle (self, session):
		"""
		Construct a new search handle for the given xesam.server.Session.
		
		NOTE: In tis implementation search handles *must* be globally unique
		"""
		self._search_counter = str (int(self._search_counter) + 1)
		return session.get_handle() + ":" + self._search_counter
	
	def get_session_from_search_handle (self, search_handle):
		"""
		Return the session handle for a given search handle
		"""
		session_handle = search_handle.split(":")[0]
		return session_handle

class ClientFactory :
	"""
	
	"""
	def __init__ (self, searchable, handle_factory, session_class, search_class):
		self._searcher = searchable
		self._handle_factory = handle_factory
		self._session_class = session_class
		self._search_class = search_class
	
	def new_session (self, **args):
		shandle = self._handle_factory.new_session_handle()
		return self._session_class (self._searcher, shandle, **args)
		
	def new_search (self, session, **args):
		"""
		Must have one of the named arguments 'query' or 'xml' set.
		See xesam.server.Search
		"""
		schandle = self._handle_factory.new_search_handle(session)
		return self._search_class (self._searcher, session, schandle, **args)

class Searcher (Searchable):
	"""
	Abstract class implementing xesam.server.Searchable with a few
	helper methods. This implementation also relies on a HandleFactory
	instance and a ClientFactory instance.
	
	Implementations on top of this class should use the gobject property
	"state" to get and set its state as described by the spec. Ie, use
	get_property and set_property.
	"""
	
	STATE = "state"
	
	# The properties listed here should not be confused with session properties
	__gproperties__ = {
		"state" : (gobject.TYPE_PYOBJECT, "State of search engine. A tuple (state_string, int_value)",
						"The state of the running search engine represented by a tuple (state_string, int_value). state_string can be IDLE, UPDATE, or FULL_INDEX", 
						gobject.PARAM_READWRITE),
	}
	
	def __init__ (self, handle_factory, client_factory):
		xesam.server.Searchable.__init__ (self)
		self._handle_factory = handle_factory
		self._factory = client_factory
		self._sessions = {}
		self._gprops = {self.STATE:[xesam.STATE_IDLE,"0"]}
		self._echo_queries = False
		self._running_searches = []
		
		# Make sure we emit a 
		self.connect ("notify::state", lambda obj, prop : self.emit ("state-changed", self.get_property(self.STATE)))
	
	def NewSession (self):
		session = self._factory.new_session ()
		self.add_session (session)
		return session.get_handle()
	
	def SetProperty (self, shandle, name, value):
		session = self.get_session(shandle)
		return session.set_property (name, value)
	
	def GetProperty (self, shandle, name):
		session = self.get_session (shandle)
		return session.get_property (name)
	
	def CloseSession (self, shandle):
		session = self.get_session (shandle)
		session.close()
		self.remove_session_by_id (shandle)
	
	def NewSearch (self, shandle, query_xml):
		session = self.get_session (shandle)
		search = self._factory.new_search (session, xml=query_xml)
		session.add_search (search)
		
		# relay signals from the search through the Searchable interface
		search.connect ("hits-added", lambda s, count : self.emit("hits-added", s.get_handle(), count))
		search.connect ("hits-removed", lambda s, hit_ids : self.emit("hits-removed", s.get_handle(), hit_ids))
		search.connect ("hits-modified", lambda s, hit_ids : self.emit("hits-modified", s.get_handle(), hit_ids))
		search.connect ("done", lambda s, : self.emit("search-done", s.get_handle()))
		
		if self._echo_queries:
			xesam.debug ("Query for %s was:\n%s" % (search.get_handle(),query_xml))
		
		return search.get_handle()
	
	def StartSearch (self, schandle):
		search = self.get_search(schandle)
		search.start()
		if schandle in self._running_searches:
			xesam.error ("Trying to start search %s, but it is already started" % schandle)
			return
		self._running_searches.append (schandle)
	
	def GetHitCount (self, schandle):
		if not schandle in self._running_searches:
			xesam.error ("Cannot GetHitCount on %s, search is not started" % schandle)
			raise UnstartedSearchException ("Cannot GetHitCount on %s, search is not started" % schandle)
		search = self.get_search(schandle)
		return search.get_hit_count()
		
	def GetHits (self, schandle, count):
		if not schandle in self._running_searches:
			xesam.error ("Cannot GetHits on %s, search is not started" % schandle)
			raise UnstartedSearchException ("Cannot GetHits on %s, search is not started" % schandle)
		search = self.get_search(schandle)
		hits = search.get_hits (count)
		xesam.debug ("Returning hits %s" % hits)
		return hits
	
	def GetHitData (self, schandle, hit_ids, fields):
		if not schandle in self._running_searches:
			xesam.error ("Cannot GetHitData on %s, search is not started" % schandle)
			raise UnstartedSearchException ("Cannot GetHitData on %s, search is not started" % schandle)
		search = self.get_search (schandle)
		return search.get_hit_data_by_id (hit_ids, fields)
	
	def CloseSearch (self, schandle):
		search = self.get_search (schandle)
		search.close()
	
	def GetState (self):
		return self.get_property (self.STATE)
	
	def get_session (self, session_handle):
		return self._sessions[session_handle]
	
	def get_search (self, search_handle):
		session = self.get_session(self._handle_factory.get_session_from_search_handle(search_handle))
		return session.get_search (search_handle)
	
	def add_session (self, session):
		"""
		@param session: a xesam.server.Session instance
		"""
		if not isinstance (session, Session):
			raise TypeError ("'session' argument must be xesam.server.Session, got: %s" % type(session))
		self._sessions [session.get_handle()] = session
	
	def remove_session (self, session):
		self._sessions.pop ([session.get_handle()])
	
	def remove_session_by_id (self, session_handle):
		self._sessions.pop (session_handle)
	
		
	def get_num_sessions (self):
		"""
		Return the total number of registered sessions across all clients
		"""
		return len (self._sessions)
	
	def do_get_property(self, prop):
		return self._gprops[prop.name]

	def do_set_property(self, prop, value):
		if prop.name == self.STATE:
			if isinstance(value,list) \
			   and len(value) == 2    \
			   and isinstance(value[0], basestring) \
			   and isinstance(value[1], basestring):
			   	v = int(value[1])
			   	if not value[0] in xesam.STATES:
			   		xesam.error ("Invalid state %s, must be one of %s. Using anyway." % (value[0], ", ".join(xesam.STATES)))
			   	if not (v >= 0 and v <= 100):
			   		xesam.error ("Invalid state value %s. Must be between 0 and 100. Using anyway." % value[1])
				self._gprops[self.STATE] = value
			else:
				raise ValueError ("Invalid value for property 'state': %s. Not a pair [str,str]" % value)
		else:
			print prop.name, prop.value_type, dir(prop), str(prop)
			raise KeyError ("Unknown property %s" % prop)
	
	def set_echo_queries (self, do_echo):
		self._echo_queries = do_echo		
		
class SearchServerStub (GenericServer):
	"""
	Stub implementation of a dbus service implementing the Xesam search api.
	All method calls are relayed to a backend implementation of SearchServerIFace.
	See dummy.DummySearchBackend for an example.	
	"""
	
	def __init__ (self, searchable):
		"""
		Relay all method calls to the SearchServerIFace specified in the constructor.
		"""
		GenericServer.__init__ (self, xesam.SEARCH_SERVICE["application"], xesam.SEARCH_SERVICE["object"])
		self.backend = searchable
		
		# A map from dbus-connection -> list of active sessions
		self._x_sessions = {}
		
		# When backend fires gobject signals relay thems as dbus signals
		self.backend.connect ("hits-added", lambda bcknd, sh, count : self.HitsAdded(sh, count))
		self.backend.connect ("hits-removed", lambda bcknd, sh, ids : self.HitsRemoved(sh, ids))
		self.backend.connect ("hits-modified", lambda bcknd, sh, ids : self.HitsModified(sh, ids))
		self.backend.connect ("search-done", lambda bcknd, sh : self.SearchDone(sh))
		self.backend.connect ("state-changed", lambda bcknd, state_info : self.StateChanged(state_info))
		
		# Listen for disconnecting clients to clean up potential unclosed sessions		
		self.dbus.connect_to_signal ("NameOwnerChanged", self._name_owner_changed)
		
	@dbus.service.method("org.freedesktop.xesam.Search",
	                         in_signature='', out_signature='s',
	                         sender_keyword='sender')
	def NewSession (self, sender):
		"""
		Sender should be the name of the dbus object doing the request
		Fx ":1.24"
		"""
		if not self._x_sessions.has_key (sender):
			self._x_sessions[sender] = []
			
		session = self.backend.NewSession()
		self._x_sessions[sender].append(session)
		
		xesam.debug("Created new session %s for object at name %s" % (session, sender))
		
		return session
	
	@dbus.service.method("org.freedesktop.xesam.Search",
	                         in_signature='ssv', out_signature='v')
	def SetProperty (self, session, name, value):
		return self.backend.SetProperty(str(session), str(name), xesam.pyvalue(value))
	
	@dbus.service.method("org.freedesktop.xesam.Search",
	                         in_signature='ss', out_signature='v')
	def GetProperty (self, session, name):
		value = self.backend.GetProperty(str(session), str(name))
		return value
	
	@dbus.service.method("org.freedesktop.xesam.Search",
	                         in_signature='s', out_signature='')
	def CloseSession (self, session):
		self.backend.CloseSession(str(session))
		xesam.debug ("Closed session %s" % session)
		
		remove_sender = None
		
		# Update our internal list of sessions
		for sender, connections in self._x_sessions.iteritems():
			for ses in connections:
				if ses == session:
					connections.remove(ses)
					if len(connections) == 0:
						remove_sender = sender
					break
		
		if remove_sender:
			self._x_sessions.pop(remove_sender)
			xesam.debug ("Sender %s disconnected" % remove_sender)
		
	
	@dbus.service.method("org.freedesktop.xesam.Search",
                         in_signature='ss', out_signature='s')
	def NewSearch (self, session, query_xml):
		schandle = self.backend.NewSearch(str(session), str(query_xml))
		xesam.debug ("Created new search %s for session %s" % (schandle, session))
		return schandle
	
	@dbus.service.method("org.freedesktop.xesam.Search",
                         in_signature='s', out_signature='')
	def StartSearch (self, search):
		return self.backend.StartSearch(str(search))
	
	@dbus.service.method("org.freedesktop.xesam.Search",
                         in_signature='s', out_signature='u')
	def GetHitCount (self, search):
		return self.backend.GetHitCount(str(search))
	
	@dbus.service.method("org.freedesktop.xesam.Search",
                         in_signature='su', out_signature='aav')
	def GetHits (self, search, count):
		return self.backend.GetHits(str(search), count)
	
	@dbus.service.method("org.freedesktop.xesam.Search",
                         in_signature='sauas', out_signature='aav')
	def GetHitData (self, search, hit_ids, properties):
		return self.backend.GetHitData(str(search), hit_ids, xesam.pyvalue(properties))
	
	@dbus.service.method("org.freedesktop.xesam.Search",
                         in_signature='s', out_signature='')
	def CloseSearch (self, search):
		xesam.debug ("Closing search %s" % search)
		return self.backend.CloseSearch(str(search))
	
	@dbus.service.method("org.freedesktop.xesam.Search",
                         in_signature='', out_signature='as')
	def GetState (self):
		return self.backend.GetState ()
	
	@dbus.service.signal("org.freedesktop.xesam.Search",
                         signature="su")
	def HitsAdded (self, search, count):
		# this is a dbus signal so no impl is needed
		pass
		
	@dbus.service.signal("org.freedesktop.xesam.Search",
                         signature="sau")
	def HitsRemoved (self, search, hit_ids):
		# this is a dbus signal so no impl is needed
		pass
		
	@dbus.service.signal("org.freedesktop.xesam.Search",
                         signature="sau")
	def HitsModified (self, search, hit_ids):
		# this is a dbus signal so no impl is needed
		pass
	
	@dbus.service.signal("org.freedesktop.xesam.Search",
                         signature="s")
	def SearchDone (self, search):
		# this is a dbus signal so no impl is needed
		pass
		
	@dbus.service.signal("org.freedesktop.xesam.Search",
                         signature="as")
	def StateChanged (self, state_info):
		# this is a dbus signal so no impl is needed
		pass
	
	def _name_owner_changed (self, name, old, new):
		if not self._x_sessions.has_key (old):
			return
		
		xesam.debug ("Connection %s closed unexpectedly" % old)
		for session in self._x_sessions[old]:
			self.CloseSession(session)

class ServiceShutdown (Exception):
	def __init__ (self):
		Exception.__init__ (self, "The ShortLivedSearcher has shut down because of inactivity")

class ShortLivedSearcher (Searcher):
	"""
	This is a special server that exits the event loop and de-registers
	from the bus when the last client exits.
	
	It can be used to create executable style search engines that are spawned
	on demand by dbus activation and exit when there's nothing to do.
	
	To use this class to create actual implementations of SearchServerIFace,
	just fill in the remaining methods.
	"""
	
	# Time with inactivity before we shut down
	TIMEOUT = 5000
	
	def __init__ (self, handle_factory, client_factory):
		Searcher.__init__ (self, handle_factory, client_factory)
		
		# Export our selves via a SearchServerStub
		self._dbus_iface = SearchServerStub(self)
		
		self._timeout_handle = gobject.timeout_add (self.TIMEOUT, self._disconnect)
	
	def start (self):
		"""
		Start a main event loop. If the search server is idle
		for TIMEOUT ms this method will issue a ServiceShutdown
		exception.
		
		You can about it manually by calling a stop(). This will not
		raise a ServiceShutdown exception.
		"""
		self._dbus_iface.start()
	
	def stop (self):
		try:
			self._disconnect()
		except ServiceShutdown:
			# This is the expected behavior
			pass
	
	def _update_timeout (self):
		if not self._timeout_handle is None:
			gobject.source_remove (self._timeout_handle)
		self._timeout_handle = gobject.timeout_add (self.TIMEOUT, self._disconnect)
	
	def _disconnect (self):
		if self.get_num_sessions () > 0:
			xesam.debug ("%s active sessions. Staying up." % self.get_num_sessions())
			self._update_timeout()
			return
		
		# There are no active sessions
		xesam.debug ("No active sessions. Shutting down.")
		self._dbus_iface.stop()		
		raise ServiceShutdown()

class UnknownFieldException (Exception):
	def __init__ (self, field):
		Exception.__init__ (self, field)

class Session :
	"""
	Default implementation of a server side Session object.
	It accepts all properties.
	"""
	def __init__ (self, searcher, session_handle):
		self._searcher = searcher
		self._handle = session_handle
		self._searches = {} # map of search_handle -> search object
		self._properties = {} # map of propname -> value (remember that value is a dbus variant and not just a string)
	
	def get_handle (self):
		return self._handle
	
	def get_property (self, prop):
		val = None
		try:
			val = self._properties[prop]
		except KeyError:
			try:
				# Return default value
				val = xesam.PROPERTIES[prop].get_value()
			except KeyError, e:
				xesam.warn ("Request of unknown session property %s on session %s with properties %s" %
								(prop, self._handle, self._properties))
				raise e
		return val

	def set_property (self, prop, val):
		prop = pyvalue(prop); val = pyvalue(val)
		self._properties[prop] = val
		return self._properties[prop]
	
	def add_search (self, search):
		"""
		Will raise a DuplicateSearchException if there's already a
		search registered with the given handle
		@param search: a xesam.server.Search instance
		"""
		if self._searches.has_key (search.get_handle()):
			raise DuplicateSearchException (search)
			
		self._searches[search.get_handle()] = search
	
	def get_search (self, search_handle):
		return self._searches[search_handle]
	
	def remove_search (self, search):
		if not search.is_closed():
			search.close()
		self._searches.pop (search.get_handle())
	
	def get_num_searches (self):
		return len (self._searches)
	
	def close (self):
		for search in self._searches.itervalues():
			if not search.is_closed():
				# to avoid chaning the dict while iterating we pass
				# detach_from_session=False to the Search.close method
				search.close(detach_from_session=False)
		
		self._searches.clear()
		
		del self._handle
		del self._searches
		del self._properties

class StopGetHits (Exception):
	pass
		

class DuplicateSearchException (Exception):
	"""
	Raised if a Session encounters two searches with the same handle
	@param search: a xesam.server.Search object
	"""
	def __init__ (self, search):
		Exception.__init__ (self, search.get_handle())

class ClosedSearchException (Exception):
	"""
	Raised if you try to read data from a closed search
	"""
	def __init__ (self, search):
		"""
		@param search: a xesam.server.Search
		"""
		Exception.__init__ (self, search.get_handle())
		
class Search (gobject.GObject):
	"""
	Server side representation of a Search object
	"""
	
	__gsignals__ = {
		"hits-added" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_INT]),
		"hits-removed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_PYOBJECT]),
		"hits-modified" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_PYOBJECT]),
		"done" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, []),
	}
	
	def __init__ (self, searchable, session, search_handle, query=None, xml=None):
		"""
		You must pass one and only one of the arguments query or xml.
		
		@param searcher: a xesam.server.Searchable
		@param session: is a xesam.server.Session instance
		@param search_handle: is the handle to use for this Search object
		@param query: a xesam.query.Query
		@param xml: a string or file like object with a xesam query in xml format
		"""
		gobject.GObject.__init__ (self)
		
		if query and xml:
			# Both args are set
			raise TypeError("One and exactly one of the named arguments 'query' and 'xml' must be present. Both set.")
		
		if xml:
			self._query = xesam.query.get_parser().parse(xml)
		elif query:
			if not isinstance (query, xesam.query.Query): raise TypeError ("query must be a xesam.query.Query, got %s" % type(query))
			self._query = query
		else:
			# None of xml and query are set
			raise TypeError ("One and exactly one of the named arguments 'query' and 'xml' must be present. None set.")
		
		self._searcher = searchable
		self._session = session
		self._handle = search_handle
		
		self._hits = HitCache()
		self._hit_queue = Queue()
		self._last_read = -1
		
		self._stopped = False
		self._closed = False
	
	def start (self):
		"""
		Start the search
		"""
		raise NotImplementedError
	
	def get_query (self):
		return self._query
	
	def get_property (self, prop):
		return self._session.get_property(prop)
	
	def get_handle (self):
		return self._handle
	
	def get_session (self):
		return self._session
	
	def get_hit_count (self):
		"""
		Return the total number of hits ever found by this search
		"""
		return len(self._hits)
	
	def get_num_ready (self):
		"""
		Return the number of hits ready to be read
		"""
		if self._closed :
			raise ClosedSearchException (self)
		return self._hit_queue.qsize ()
	
	def add_new_hit (self, fields, values):
		"""
		Add hits to this Search. Hit objects will be created from the
		data passed to this method.
		
		It is possible to pass more fields than requested via hit.fields.
		All hit data will be cached and a hit matching the requested hit.fields
		will be generated.
		
		@param fields: a list of field names
		@param values: a list of values corresponding to the field names
		"""
		if self._closed :
			raise ClosedSearchException (self)
			
		self._hits.put_data(len(self._hits), fields, values)
		self._hit_queue.put_nowait(Hit(*self._hits.get(len(self._hits) - 1, fields)))
	
	def update_hits (self, hits):
		"""
		Used by the parent Searcher to add additional data for hits. Mainly used to
		support for example GetHitData calls.
		The Searcher.GetHitData method calls into Search.get_hit_data where  as much
		as possible is retrieved from the cache. Any missing data will be retrieved
		via this method.
		
		The returned structure is a dict of dicts:
		  {hit_id_1 : {field1:value1, field2:value2,...}, ...}
		"""
		xesam.debug ("[Search:%s] Got hit data update for %s hits" % (self._handle, len(hits)))
		
		for hit in hits:
			if hit.get_id() < 0 or hit.get_id() >= len(self._hits) :
				xesam.error ("Received update for invalid hit id '%s'" % hit.get_id())
				return
			self._hits.put_data(hit.get_id(), hit.get_data().iterkeys(), hit.get_data().itervalues())
	
	def get_hits (self, num_hits):
		xesam.debug ("[Search:%s] Got request for %s hits" % 
											(self._handle, num_hits))
		if self._closed :
			raise ClosedSearchException (self)
			
		result = []
		if not self._stopped:
			for i in range(num_hits):
				hit = self._hit_queue.get()
				if isinstance (hit, StopGetHits):
					# We have been requested to stop blocking
					break
				result.append(hit.get_values())
		else:
			for i in range(num_hits):
				try:
					hit = self._hit_queue.get_nowait()
					if isinstance (hit, StopGetHits):
						# We have been requested to stop blocking
						break
					result.append(hit.get_values())
				except Empty:
					break
		
		xesam.debug ("[Search:%s] Read %s hits" % (self._handle, len(result)))
		return result
	
	def remove_hits (self, hits):
		"""
		Invalidate a list of hits. This is only meaningful on live 
		searches.
		@param hits: a list of xesam.server.Hit instances
		"""
		if self._closed :
			raise ClosedSearchException (self)
		for hit in hits:
			self._hits.delete (hit.get_id())
	
	def remove_hits_by_id (self, hit_ids):
		if self._closed :
			raise ClosedSearchException (self)
		for hit_id in hit_ids:
			self._hits.delete(hit_id)

		
	def get_hit_data (self, hits, fields):
		"""
		Get additional (or renewed) metadata for a list of Hits.
		The returned list of Hit objects have the same ids (in the same order)
		as the ones provided in the arguments.
		"""
		if self._closed :
			raise ClosedSearchException (self)
		hit_ids = [hit.get_id() for hit in hits]
		return self.get_hit_data_by_id(hit_ids, fields)
	
	def get_hit_data_by_id (self, hit_ids, fields):
		"""
		Get additional (or renewed) metadata from a set of hits
		given their ids.
		
		If you want to intercept calls the the Searcher's GetHitData
		and update the HitCache accordingly, override this method
		and return via the native implementation of Search.get_hit_data_by_id()
		in the end of your method like
		
		my_update_the_cache_with_new_fields()
		
		return Search.get_hit_data_by_id(hit_ids, fields)
		"""
		if self._closed :
			raise ClosedSearchException (self)

		hits = []
		for hit_id in hit_ids:
			hit_id, f, values = self._hits.get (hit_id, fields)
			hits.append (values)
		return hits
	
	def get_hit_cache (self):
		"""
		Return a reference to the HitCache used to store hit data for this search.
		"""
		return self._hits
	
	def stop (self):
		"""
		Stop the search and signal that no more hits will be comming in.
		This means that any call to get_hits() will return immediately.
		
		Calling this method will trigger a "done" signal.
		
		This should for example be used when the entire index has been
		scanned and xesam.SESSION_SEARCH_LIVE is false.
		"""
		if self._closed :
			raise ClosedSearchException (self, "Cannot stop search %s. The search is already closed" % self._handle)
		self._hit_queue.put_nowait (StopGetHits())
		self._stopped = True
		
		self.emit ("done")
	
	def is_closed (self):
		return self._closed
	
	def close (self, detach_from_session=True):
		"""
		Clean up and free all hits an related data. The Search instance
		is invalid after this call.
		
		This call will remove the Search object from the parent Session
		unless you add the detach_from_session=False named argument.
		"""
		if self._closed :
			# it is ok to close multiple times
			return
		
		self._closed = True	
		
		# Detach from session
		if detach_from_session:
			self._session.remove_search (self)
		
		# clean up
		self._hit_queue.put_nowait (StopGetHits())

class HitCache :
	"""
	Used by xesam.server.Search objects to store field data.
	Warning: 	Current implementation is heavily un-optimised.
	"""
	def __init__ (self):
		self._hits = {}
	
	def put_data (self, hit_id, field_names, field_data):
		assert (len(field_names) == len(field_data))
		
		field_map = None
		
		if not self._hits.has_key(hit_id):
			field_map = {}
			self._hits[hit_id] = field_map
		else:
			field_map = self._hits[hit_id]
		
		for i in range(len(field_names)):
			field_map [field_names[i]] = field_data[i]
	
	def put_hits (self, hits):
		for hit in hits:
			if not self._hits.has_key (hit.get_id()):
				xesam.error ("Trying to update data for uncached hit %s" % hit.get_id())
			else:
				self._hits[hit.get_id()].update (hit.get_data())
		
	
	def get (self, hit_id, field_names):
		if not self._hits.has_key(hit_id):
			xesam.error ("Hit '%s' not cached. Returning empty hit data." % hit_id)
			return Hit (hit_id, field_names, ["" for i in field_names])
		
		hit_data = self._hits[hit_id]
		values = []
		for i in range(len(field_names)):
			if hit_data.has_key(field_names[i]):
				values.append(hit_data[field_names[i]])
			else:
				values.append(None)
				xesam.error ("Field '%s' not cached for hit '%s'. Returning None." % (field_names[i], hit_id))
		
		return (hit_id, field_names, values)
		
	def delete (self, hit_id):
		if not self._hits.has_key (hit_id):
			xesam.error ("Trying to delete uncached hit %s" % hit_id)
			return
		self._hits [hit_id] = None
	
	def has_hit (self, hit_id):
		return self._hits.has_key (hit_id)
		
	def hit_has_field (self, hit_id, field):
		if not self._hits.has_key (hit_id):
			return False
		if (not self._hits[hit_id].has_key(field)) or (self._hits[hit_id][field] is None):
			return False
		return True
	
	def __len__ (self):
		return len(self._hits)

class Hit :
	"""
	You can look up metadata values as in a dict by looking up
	directly on this object. Like 
	
		uri = my_hit["uri"]
		
	"""
	def __init__ (self, hit_id, fields, values):
		self._id = hit_id
		self._data = {}
		self._values = values
		
		# Map property names to their corresponding values
		for i in range(len(fields)):
			self._data[fields[i]] = xesam.pyvalue(values[i])		
		
	def __getitem__ (self, key):
		return self._data[key]
	
	def __iter__ (self):
		return self._data.iteritems()
	
	def get_values (self):
		return self._values
	
	def get_id (self):
		return self._id
	
	def get_data (self):
		return self._data
	
	def __str__ (self):
		result = "[Hit:%s]" % self.get_id()
		for field,val in self:
			result += "\n%s=%s" % (field,val)
		return result
