import gtk, gobject, pango

from xesam.client import *

class HitPagerModel (gtk.TreeStore):
	"""
	
	"""
	
	__gproperties__ = {
		"page-size" : (gobject.TYPE_UINT, "Number of hits per page",
						"Number of hits to display per page",
						0, 100, 15, gobject.PARAM_READWRITE),
		"page" : (gobject.TYPE_UINT, "Page number",
						"Page number of currently displayed page",
						1, 100, 1, gobject.PARAM_READWRITE),
		"has-next-page" : (gobject.TYPE_BOOLEAN, "True if there is another page of hits",
						"Whether or not there are hits available on the next page",
						False, gobject.PARAM_READABLE),
		"has-previous-page" : (gobject.TYPE_BOOLEAN, "True if you can go to the previous page",
						"Whether or not you can go to the previous page",
						False, gobject.PARAM_READABLE),
	}
	
	def __init__ (self, session):
		gtk.TreeStore.__init__ (self, gobject.TYPE_PYOBJECT) # We store xesam.client.Hit objects
		self._session = session
		self._search = None
		self._search_closed_handle = 0
		self._hits_added_handle = 0
		self._hits_modified_handle = 0
		self._hits_removed_handle = 0
		self._found_hits = 0
		self._page_size = 15
		self._page = 1
		self._has_next_page = False
		self._has_prev_page = False
		self._current_page = 1
		self._first_id = (self._page - 1) * self._page_size
		self._last_id = self._page * self._page_size - 1
		self._hit_fields = session.get_property ("hit-fields")
		self._hits = []
		
		session.connect ("notify::hit-fields", self._warn_hit_fields_changed)
		session.connect ("closed", self._on_session_closed)
		self.connect ("notify::page", lambda x, y : self._on_page_changed())
		
		
	def set_search (self, search):
		self.reset()
		self._search_closed_handle = search.connect ("closed", self._on_search_closed)
		self._hits_added_handle = search.connect ("hits-added", self._on_hits_added)
		self._hits_modified_handle = search.connect ("hits-modified", self._on_hits_modified)
		self._hits_removed_handle = search.connect ("hits-removed", self._on_hits_removed)
		self._search = search
	
	def clear (self):
		gtk.TreeStore.clear(self)
	
	def reset (self):
		if self._search_closed_handle:
			gobject.source_remove (self._search_closed_handle)
		if self._hits_added_handle:
			gobject.source_remove (self._hits_added_handle)
		if self._hits_removed_handle:
			gobject.source_remove (self._hits_removed_handle)
		if self._hits_modified_handle:
			gobject.source_remove (self._hits_modified_handle)
			
		if self._search:
			self._search = None
			
		self._found_hits = 0
			
		self.clear ()
	
	def next_page (self):
		self.set_property ("page", self.get_property ("page") + 1)
	
	def previous_page (self):
		self.set_property ("page", self.get_property ("page") - 1)
	
	def _on_search_closed (self, search):
		xesam.debug ("Caught search 'closed' signal. Cleaning up.")
		self.reset()
	
	def _warn_hit_fields_changed (self, session):
		xesam.error ("%s does not support changing the 'hit-fields' property on a running session. Behavior will be erratic from mow on." % self.__class__)
		
	def _on_session_closed (self, session):
		xesam.debug ("%s's session was closed - resetting %s" % self)
		self.reset()
	
	def _on_hits_added (self, search, count):
		xesam.debug ("%s hits added on search %s" % (count, self._search))
		self._found_hits += count
		
		missing = self._last_id - len(self._hits) + 1
		
		# If we have everything we need, leave the hits on the server for now
		if missing <= 0:
			return
		
		to_read = min (missing, count)
		xesam.debug ("Reading %s hits out of %s available, and %s missing to fill page" 
						% (to_read, self._found_hits - len(self._hits), missing))
		new_hits = self._search.get_hits (to_read)
		xesam.debug ("Read %s hits" % len(new_hits))
		self._hits.extend (new_hits)
		
		assert (len(new_hits) == to_read)
		
		# When we reach this point it is always because the hits where needed for display,
		# so we can calmly update the tree model
		for hit in new_hits:
			self.append (None, [hit])
		
		self._update_page_availability()
	
	def _on_hits_modified (self, search, hit_ids):
		xesam.debug ("%s hits modified on search %s" % (len(hit_ids), self._search))
		to_retrieve = []
		
		for hit_id in hit_ids:
			if hit_id > len(self._hits):
				xesam.debug ("Hit '%s' changed, but has not yet been retrieved. Ignoring." % hit_id)
			else:
				to_retrieve.append (hit_id)
				
		updated_hits = self._search.get_hit_data_from_ids (to_retrieve, self._hit_fields)
		
		# Update our internal hit cache
		for hit in updated_hits:
			self._hits[hit.get_id()] = hit
			
			# Update visible rows
			if hit.get_id() <= self._last_id and hit.get_id () >= self._first_id:
				iter = self.iter_nth_child (hit.get_id() - self._first_id)
				self.emit ("row-changed", self.get_path(iter), iter)
			
		
		
	def _on_hits_removed (self, search, hit_ids):
		xesam.debug ("%s hits removed on search %s" % (len(hit_ids), self._search))
		
		update_view = False
		
		for hit_id in hit_ids:
			if hit_id > len(self._hits):
				xesam.debug ("Hit '%s' removed, but has not yet been retrieved. Ignoring." % hit_id)
			else:
				if hit_id < self._last_id and hit_id >= self._first_id:
					update_view = True
				self._hits[hit_id] = None
		
		if update_view:
			self._on_page_changed()

	def _update_first_last_id (self):
		self._first_id = (self._page - 1) * self._page_size
		self._last_id = self._page * self._page_size - 1
	
	def _update_page_availability (self):
		if self._page == 1:
			self._has_prev_page = False
		else:
			self._has_prev_page = True			
		
		if self._found_hits > self._last_id:
			self._has_next_page = True
		else:
			self._has_next_page = False
			
		self.notify ("has-previous-page")
		self.notify ("has-next-page")	
				
	def _on_page_changed (self):
		self.clear ()		
		self._update_first_last_id ()
		
		num_ready = len(self._hits) - (self._last_id + 1)
		missing = abs(min(0, num_ready)) # If num_ready < 0, that is the number of hits we miss
		can_read = self._found_hits - len(self._hits)
		to_read = min (missing, can_read)
		
		if to_read > 0:
			xesam.debug ("Reading %s hits out of %s available, and %s missing to fill page" 
						% (to_read, self._found_hits - len(self._hits), missing))
			new_hits = self._search.get_hits (to_read)
			xesam.debug ("Read %s hits" % len(new_hits))
			self._hits.extend(new_hits)
		
		last_display = min (self._last_id, len(self._hits) - 1)
		
		xesam.debug ("Changing page to %s, showing hits with id range [%s, %s]"
						% (self._page, self._first_id, self._last_id))
		
		## FIXME: This does not take removed hits into account (hint: they are stored as None in self._hits)
		
		for hit in self._hits[self._first_id:last_display+1]:
			self.append (None, [hit])
			
	def do_get_property(self, prop):
		if prop.name == "page-size":
			return self._page_size
		elif prop.name == "page":
			return self._page
		elif prop.name == "has-next-page":
			return self._has_next_page
		elif prop.name == "has-previous-page":
			return self._has_prev_page
		else:
			xesam.error ("Request to get unknown property '%s'" % prop)
			return None
		
	def do_set_property(self, prop, value):
		if prop.name == "page-size":
			if value < 1:
				xesam.error ("non-positive page-size requested '%s', using default 15", value)
				self._page_size = 15
			else:
				self._page_size = value
		elif prop.name == "page":
			if value < 1:
				xesam.error ("non-positive page '%s' requested, using default 1", value)
				self._page = 1
			else:				
				self._page = value			
		else:
				xesam.error ("Request to set unknown property '%s'" % prop)
				return
				
		# Update properties with indirect values
		self._update_first_last_id ()
		self._update_page_availability()

class HitView (gtk.TreeView):
	"""
	
	"""
	def __init__ (self, model):
		gtk.TreeView.__init__ (self, model)
		self._hit_msg = "Empty hit message"
		
		self.set_property ("headers-visible", False)
		self.set_enable_search (False)
		
		        
		hit_msg_cell = gtk.CellRendererText ()
		hit_msg_cell.set_property ("ellipsize", pango.ELLIPSIZE_END)
		
		hit_col = gtk.TreeViewColumn ("Hits")
		hit_col.pack_start (hit_msg_cell)
		hit_col.set_cell_data_func (hit_msg_cell, self._get_hit_msg_for_cell)
		self.append_column (hit_col)		
        
	def set_hit_message (self, msg):
		"""
		You can escape hit fields like %(xesam:url)s and use Pango markup.
		"""
		self._hit_msg = msg
	
	def _get_hit_msg_for_cell (self, column, cell, model, iter, data=None):
		hit = model[iter][0]
		
		cell.set_property ("markup", self._hit_msg % hit)

